<?php

namespace NoccyLabs\Screen;

use Stringable;

class Buffer
{
    private ?string $buffer = null;

    private int $translateX;

    private int $translateY;

    private static $theme = [];

    public function __construct(int $translateX=1, int $translateY=1)
    {
        $this->translateX = $translateX;
        $this->translateY = $translateY;
    }

    /**
     * Define a theme element
     * 
     * 
     * @param null|string Theme element to use
     * @param null|int|string|array The foreground color
     * @param null|int|string|array The background color
     * @param null|StyleEnum[] Styles
     */
    public static function theme(string $theme, int|string|array|null $fg=null, int|string|array|null $bg=null, array|null $style=null)
    {
        self::$theme[$theme] = [
            'fg' => $fg,
            'bg' => $bg,
            'style' => $style,
        ];
    }

    public function at(int $line, int $col, bool $absolute=false): self
    {
        if ($absolute) {
            $this->buffer .= "\e[".($line+1).";".($col+1)."H";
        } else {
            $this->buffer .= "\e[".($line+$this->translateY).";".($col+$this->translateX)."H";
        }
        return $this;
    }

    /**
     * Set the Graphic Rendition (according to ANSI nomenclature) - Simply set the output style.
     * 
     * 
     * @param null|string Theme element to use
     * @param null|int|string|array The foreground color
     * @param null|int|string|array The background color
     * @param null|int[] Styles
     */
    public function sgr(?string $theme=null, int|string|array|null $fg=null, int|string|array|null $bg=null, array|null $style=null): self
    {
        if ($theme && array_key_exists($theme, self::$theme)) {
            $fg = ($fg!==null) ? $fg :self::$theme[$theme]['fg'];
            $bg = ($bg!==null) ? $bg : self::$theme[$theme]['bg'];
            $style = ($style!==null) ? $style : self::$theme[$theme]['style'];
        }
        $b = [];
        if ($style!==null) $b[] = join(";", $style);
        if ($fg!==null) $b[] = $this->makeAnsiColor($fg, true);
        if ($bg!==null) $b[] = $this->makeAnsiColor($bg, false);
        $this->buffer .= "\e[".join(";", $b)."m";

        return $this;
    }

    public function write(string|null $content): self
    {
        $this->buffer .= $content;
        return $this;
    }

    public function clear(int $lines, int $cols): self
    {
        for ($l = 0; $l < $lines; $l++) {
            $this->at($l, 0)->write(str_repeat(" ", $cols));
        }

        return $this;
    }

    private function makeAnsiColor(int|string|array $color, bool $foreground): string
    {
        if (is_int($color)) {
            if ($color < 16) {
                // output 16-color code
                $cv = $color % 8;
                $ci = $color >= 8;
                return (($ci?($foreground?90:100):($foreground?30:40)) + $cv);
            } else {
                // output 256-color code
                return ($foreground?38:48).";5;".$color;
            }
        }
        return '';
    }

    public function append(Buffer $buffer): self
    {
        $this->buffer .= $buffer->flush(true);
        return $this;
    }

    public function flush(bool $return=false)
    {
        $r = $this->buffer;
        $this->buffer = "";
        if ($return) {
            return $r;
        }
        echo $r;
    }
}

